#ifdef DREAMCAST
#include <kos.h>
extern uint8 romdisk[];
KOS_INIT_FLAGS(INIT_DEFAULT);
KOS_INIT_ROMDISK(romdisk);
#include <SDL.h>
#include <SDL_dreamcast.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <SDL.h>

#include "zx4all.h"
#include "z80intrf.h"
#include "computer.h"
#include "snapshot.h"
#include "characters.h"
#include "menus.h"
#include "sound.h"
#include "tape.h"
#include "microdrive.h"
#include "debug.h"

#include "vkbd/vkbd.h"
#include "vtape/vtape.h"

#ifndef USE_RAZE
Z80 procesador;
#else
unsigned __z80_last_cycles;
#endif

#include "menu.h"

char zx4all_image_file[1024];
char zx4all_tape_file[1024];
int emulating=0;
char salir=0;

unsigned long long DEBUG_ACTUAL_TIMESLICE=0;

#ifndef DEBUG_Z80_STATS
#if !defined(DEBUG_START) || DEBUG_START==0
#define DEBUG_START 0
int DEBUG_AHORA=1;
#else
int DEBUG_AHORA=0;
#endif

#ifndef DEBUG_END
#define DEBUG_END (DEBUG_START+10)
#endif
#endif

struct computer cmpt;
SDL_Surface *screen;
unsigned char *sound[NUM_SNDBUF];
/*
char path_snaps[2049];
char path_taps[2049];
char path_mdrs[2049];
*/

#ifndef DREAMCAST
static char original_path[1024];
#endif

extern int closing_sound;
extern int mainMenu_sound;
extern int menuOpt_mode;
extern int menuOpt_joy;
extern int menuOpt_ay;
extern int menuOpt_intr;

void setChanges(void)
{
	if (cmpt.issue<=2 && cmpt.mode128k==0)
	{
		menuOpt_mode=0;
		menuOpt_ay=0;
	}
	else
	if (cmpt.mode128k==0)
	{
		menuOpt_mode=1;
		menuOpt_ay=0;
	}
	else
	if (cmpt.mode128k==1)
	{
		menuOpt_mode=2;
		menuOpt_ay=1;
	}
	else
	if (cmpt.mode128k==2)
	{
		menuOpt_mode=3;
		menuOpt_ay=1;
	}
	else
	{
		menuOpt_mode=4;
		menuOpt_ay=1;
	}

}

void getChanges(void)
{
#ifndef NOSOUND
	if (!mainMenu_sound)
	{
		sound_type=1;
		closing_sound=1;
	}
	else
		sound_type=0;
#endif

	switch (menuOpt_mode)
	{
		case 0:
			if (cmpt.issue>2 || cmpt.mode128k!=0)
			{
				cmpt.issue=2;
				cmpt.mode128k=0;
				cmpt.ay_emul=menuOpt_ay=0;
				ResetComputer();
			}
			break;
		case 1:
			if (cmpt.issue!=3 || cmpt.mode128k!=0)
			{
				cmpt.issue=3;
				cmpt.mode128k=0;
				cmpt.ay_emul=menuOpt_ay=0;
				ResetComputer();
			}
			break;
		case 2:
			if (cmpt.issue!=3 || cmpt.mode128k!=1)
			{
				cmpt.issue=3;
				cmpt.mode128k=1;
				cmpt.ay_emul=menuOpt_ay=-1;
				ResetComputer();
			}
			break;
		case 3:
			if (cmpt.issue!=3 || cmpt.mode128k!=2)
			{
				cmpt.issue=3;
				cmpt.mode128k=2;
				cmpt.ay_emul=menuOpt_ay=-1;
				ResetComputer();
			}
			break;
		case 4:
			if (cmpt.issue!=3 || cmpt.mode128k!=3)
			{
				cmpt.issue=3;
				cmpt.mode128k=3;
				cmpt.ay_emul=menuOpt_ay=-1;
				cmpt.mdr_active=menuOpt_intr=0;
				ResetComputer();
			}
			break;
	}
	cmpt.ay_emul=menuOpt_ay;
	cmpt.joystick=menuOpt_joy;
	if (cmpt.mdr_active!=menuOpt_intr)
	{
		cmpt.mdr_active=menuOpt_intr;
/*
		if(cmpt.mode128k!=3)
			ResetComputer();
*/
	}
}

void load_rom(char type) {
	FILE *fichero;
	int size;
	char *fname=(char *)calloc(1,1024);
	int fname_len;
#ifndef DREAMCAST
	strcpy(fname,original_path);
#endif
	strcat(fname,ROM_PATH_PREFIX);
	fname_len=strlen(fname);

	switch(type) {
	case 0:
		strcat(fname,"/spectrum.rom");
		fichero=fopen(fname,"r");
		size=16384;
		break;
	case 1:
		strcat(fname,"/128.rom");
		fichero=fopen(fname,"r");
		size=32768;
	break;
	case 2:
		strcat(fname,"/plus2.rom");
		fichero=fopen(fname,"r");
		size=32768;
	break;
	case 3:
		strcat(fname,"/plus3.rom");
		fichero=fopen(fname,"r");
		size=65536;
	break;
	default:
		fichero=NULL;
		size=0;
	}
  
	if(fichero==NULL) {
		printf("Can't open ROM file\n");
		exit(1);
	}
	fread(cmpt.memoria,size,1,fichero);
	fclose(fichero);
  
	fname[fname_len]=0;
	strcat(fname,"/if1-v2.rom");
	fichero=fopen(fname,"r"); // load Interface1 ROM
	if(fichero==NULL) {
		printf("Can't open Interface1 ROM file\n");
		exit(1);
	}
	
	fread(cmpt.shadowrom,8192,1,fichero);
  	fclose(fichero);
	free(fname);
}


static void init_sdl(void) {
  Uint32 flags=SDL_INIT_VIDEO;
#ifndef NOSOUND
  flags|=SDL_INIT_AUDIO;
#endif
#ifndef AUTO_EVENTS
  flags|=SDL_INIT_JOYSTICK;
#endif
  int retorno=SDL_Init(flags);
  if(retorno!=0) {
    printf("Can't initialize SDL library. Exiting\n");
    exit(1);
  }
#if defined(DREAMCAST) && defined(DEBUG_ZX4ALL)
  SDL_DC_ShowAskHz(SDL_FALSE);
#endif
}


static void init_screen(void) {
  // screen initialization
  Uint32 flags=SDL_HWSURFACE | SDL_DOUBLEBUF; // |SDL_HWPALETTE;
#ifdef DREAMCAST
  flags|=SDL_FULLSCREEN;
#endif

  screen=SDL_SetVideoMode(320,240,16,flags);
  if(screen==NULL) {
    printf("Can't assign SDL Surface. Exiting\n");
    exit(1);
  }

#ifndef DREAMCAST
  if(SDL_MUSTLOCK(screen)) {
    cmpt.mustlock=1;
    SDL_LockSurface(screen);
  } else
    cmpt.mustlock=0;
#endif
}

static void init_sound(void)
{
	int i;

	cmpt.channels=DEFAULT_SOUND_CHANNELS;
	cmpt.freq=DEFAULT_SOUND_FREQ;
#ifndef NOSOUND
	if(sound_init(DEFAULT_SOUND_FREQ,DEFAULT_SOUND_BUFLEN)!=0) {
#ifndef DREAMCAST
#ifndef WIN32
	  sleep(2);
#endif
	  if (sound_init(DEFAULT_SOUND_FREQ,DEFAULT_SOUND_BUFLEN)!=0)
#endif
          {
	  	printf("Can't initialize sound. Exiting. (Try with 'nosound')\n");
	  	exit(1);
	  }
	}
#else
	// NO SOUND ...
	cmpt.sign=0;
	cmpt.format=0;
	cmpt.channels=1;
	cmpt.freq=48000;
	cmpt.buffer_len=4800; // will wait 1/10 second

#endif
	for(i=0;i<NUM_SNDBUF;i++) {
		sound[i]=(unsigned char *)calloc(cmpt.buffer_len+4,cmpt.channels*2);
	}
#ifdef SOUND_IOCTL
	if(cmpt.format)
		cmpt.increment=2*cmpt.channels;
	else
#endif
		cmpt.increment=1*cmpt.channels;

	cmpt.tst_sample=3500000/cmpt.freq;
	set_volume(70);
	sound_setup();
}

void end_system() {
	
	sound_close();
	if(cmpt.tap_file!=NULL)
		gzclose(cmpt.tap_file);
#ifndef DREAMCAST
	if(cmpt.mustlock)
		SDL_UnlockSurface(screen);
#endif
	SDL_Quit();
}

static void setup_sdl(void)
{
#if !defined(DREAMCAST) && !defined(DEBUG_ZX4ALL)
	atexit(end_system);
#endif
#ifndef AUTO_EVENTS
	SDL_JoystickEventState(SDL_ENABLE);
	SDL_JoystickOpen(0);
	SDL_JoystickOpen(1);
#ifdef DREAMCAST
	SDL_DC_EmulateKeyboard(SDL_TRUE);
#endif
#else
	SDL_EventState(SDL_JOYAXISMOTION,SDL_IGNORE);
	SDL_EventState(SDL_JOYBALLMOTION,SDL_IGNORE);
	SDL_EventState(SDL_JOYHATMOTION,SDL_IGNORE);
	SDL_EventState(SDL_JOYBUTTONDOWN,SDL_IGNORE);
	SDL_EventState(SDL_JOYBUTTONUP,SDL_IGNORE);
#endif
	SDL_EventState(SDL_ACTIVEEVENT,SDL_IGNORE);
	SDL_EventState(SDL_MOUSEMOTION,SDL_IGNORE);
	SDL_EventState(SDL_MOUSEBUTTONDOWN,SDL_IGNORE);
	SDL_EventState(SDL_MOUSEBUTTONUP,SDL_IGNORE);

#ifdef DREAMCAST
	SDL_EventState(SDL_QUIT,SDL_IGNORE);
#endif
	SDL_EventState(SDL_SYSWMEVENT,SDL_IGNORE);
	SDL_EventState(SDL_VIDEORESIZE,SDL_IGNORE);
	SDL_EventState(SDL_USEREVENT,SDL_IGNORE);
	SDL_ShowCursor(SDL_DISABLE);
	vkbd_init();
	vtape_init();
}


int main(int argc,char *argv[]) {

	int bucle,tstados;
	unsigned short PC;

#ifndef DREAMCAST
	getcwd(original_path,1024);
	strcat(original_path,"/");
#endif
	puts("MAIN!!!!");

	zx4all_image_file[0]=0;

	init_sdl();
	init_screen();
	init_sound();
	setup_sdl();

	init_text(1);
	run_mainMenu();
	quit_text();
//	vkbd_init();
	zx4all_pause_music();
	zx4all_resume_sound();

	computer_init();
	register_screen(screen);
	cmpt.interr=0;

	cmpt.screenbuffer=(unsigned char*)cmpt.screen->pixels;
	cmpt.screen_width=cmpt.screen->w;

	// assign initial values for PATH variables
/*
	strcpy(path_snaps,SNAP_PATH_PREFIX "/");
	strcpy(path_taps,TAP_PATH_PREFIX "/");
	strcpy(path_mdrs,MDR_PATH_PREFIX "/");
*/
	cmpt.current_tap[0]=0;

	// assign random values to the memory before start execution

	for(bucle=0;bucle<196608;bucle++)
		cmpt.memoria[bucle]=(unsigned char) rand();

	cmpt.tap_file=NULL;
	
	// we filter all the events, except keyboard events

puts("Reseting ..."); fflush(stdout);
	ResetComputer();
puts("Z80 Init ..."); fflush(stdout);
	_z80_init();
#ifdef AUTO_SNAPSHOT
	load_z80(SNAP_PATH_PREFIX "/snap/ejemplo.z80");
#else
	if (zx4all_image_file[0])
	{
		load_z80((char *)&zx4all_image_file[0]);
		zx4all_image_file[0]=0;
	}
	if (zx4all_tape_file[0])
	{
		tape_select((char *)&zx4all_tape_file[0]);
		zx4all_tape_file[0]=0;
	}
#endif
  
//	sleep(1);

	clean_screen();

	microdrive_init();

	getChanges();
	puts("Let's go ..."); fflush(stdout);

	zx4all_remap_keys();
	while(!salir) {
#ifndef DEBUG_Z80
		PC=_z80_emulate(200);
#else
		{
			unsigned actual=0;
//			while(actual<200)
			for(actual=0;actual<34;actual++)
			{
#ifndef DEBUG_Z80_STATS
				dbg("/-");
				if (DEBUG_AHORA)
					debugZ80(0);
#endif
				PC=_z80_emulate(1);
#ifdef DEBUG_Z80_STATS
				debugZ80_GetStatitics();
#else
				dbg(" -/");
#endif
//				actual+=_z80_get_last_cycles();
			}
		}
#endif
#ifdef DEBUG_ZX4ALL
		DEBUG_ACTUAL_TIMESLICE++;
//		dbgf("ts=%i\n",((unsigned)DEBUG_ACTUAL_TIMESLICE));
#ifndef DEBUG_Z80_STATS
		if (DEBUG_ACTUAL_TIMESLICE>DEBUG_START)
			DEBUG_AHORA=1;
		if (DEBUG_ACTUAL_TIMESLICE>DEBUG_END)
		{
			puts("DEBUG END!!!"); fflush(stdout);
#ifdef DREAMCAST
			arch_reboot();
//			malloc(20*1024*1024);
#else
			exit(0);
#endif
		}
#endif
#endif
		
		/* if PC is 0x0556, a call to LD_BYTES has been made, so if
		FAST_LOAD is 1, we must load the block in memory and return */
/*
		if((!cmpt.mdr_paged)&&(PC==0x0556) && (cmpt.tape_fast_load==1)&&(cmpt.tape_file_type==TAP_TAP)) {
			if(cmpt.tap_file!=NULL)
				fastload_block(cmpt.tap_file);
		}
*/
		
		/* if PC is 0x04C2, a call to SA_BYTES has been made, so if
		we want to save to the TAP file, we do it */
/*		
		if((!cmpt.mdr_paged)&&(PC==0x04C2)&&(cmpt.tape_write==1)&&(cmpt.tape_file_type==TAP_TAP)) {
			if(cmpt.tap_file!=NULL)
				save_file(cmpt.tap_file);
		}
*/
			/* if cmpt.mdr_paged is 2, we have executed the RET at 0x0700, so
		we have to return to the classic ROM */
/*		
		if(cmpt.mdr_paged==2)
			cmpt.mdr_paged=0;
*/
		
		/* if PC is 0x0008 or 0x1708, and we have a microdrive, we have to page
		the Interface 1 ROM */
/*		
		if(((PC==0x0008)||(PC==0x1708))&&(cmpt.mdr_active))
			cmpt.mdr_paged = 1;
*/
		
		/* if PC is 0x0700 and we have a microdrive, we have to unpage
		the Interface 1 ROM after the last instruction */
/*		
		if((PC==0x0700)&&(cmpt.mdr_active))
			cmpt.mdr_paged = 2;
*/

#ifndef DEBUG_Z80
		tstados=_z80_get_last_cycles();
/*
		if(!tstados)
			tstados=4; // if we are in halt, we execute virtual NOPs
		else
		if(tstados<0) {
			printf("Error ciclado %i\n",tstados);
			exit(1);
		}
*/
		if (tstados<=0)
			tstados=4;
#else
		tstados=200;
#endif
		emulate(tstados); // execute the whole hardware emulation
		if(cmpt.interr==1) {
			register unsigned char be=bus_empty();
#if defined(DEBUG_Z80) && !defined(DEBUG_Z80_STATS)
			dbgf("INT %.2X\n",be);
#endif
			_z80_interrupt(be);
			cmpt.interr=0;
		}
	}
#if !defined(DREAMCAST) && !defined(DEBUG_ZX4ALL)
	end_system();
#endif
	return 0;
}
